// ----------------------------------------------------------------------- //
// _file.cpp	file management and stuff
//
// last edited:	2001-05-11 by Bjrn Paetzel
// 
// Copyright (C) 2001 by Bjrn Paetzel <kolrabi@gmx.de>
// ----------------------------------------------------------------------- //

// ----------------------------------------------------------------------- //
// notes
/*

 - only one pak can be loaded at any time... gotta change that some time



 - .pak format is as follows:

 offset		size (in byte)			description

 0				4						numver of files in pak
 4				136 * filecount			header entries for files as follows

 where n is the number of the file ( 0 <= n < filecount )
 
 4+136*n		128						name of file n
 4+136*n+128	4						offset of file n in bytes from beginning
										of the header so offset in bytes from
										beginning of the file is 4 + offset 
 4+136*n+132	4						the size of file n

 after this the actual data follows

 4+offset(n)	size(n)					data

 ( suggestion: add a magic? and a version field )



 - there are file buffer access routines in this file that are supposed to
   make it easier to read from buffers like from files


 - any file loaded must be freed!


 - in _fgetpixel_frombuf 16 bit pixels are interpreted as they
   are in targa files (5/5/5):

 Byte 1   Byte 2		where A is alpha and R, G and B are
 ARRRRRGG GGGBBBBB		red green and blue values respectively

*/



// ----------------------------------------------------------------------- //
// includes

#include	"stdio.h"
#include	"_shared.h"

// ----------------------------------------------------------------------- //


// ----------------------------------------------------------------------- //
// variables

long		_pak_filecount	= 0;	// number of files in pak
fileentry_t	*_pak_header	= NULL;	// pointer to the header of the loaded pak
file_t		*_pak_file		= NULL;

// ----------------------------------------------------------------------- //



// ----------------------------------------------------------------------- //
// these functions are for file access
 


// loads a file from disk into a buffer

file_t *_file_loadfromdisk			// returns a pointer to the buffer
	( 
		char *filename 				// (in)  name of the file to load
	)
{
	char	*b = NULL;				// the buffer
	FILE	*f;						// the file
	fpos_t	s;						// for file size
	file_t	*file;					// the file descriptor

	file = (file_t*)malloc(sizeof(file_t));
	if (!file) return NULL;

	file->pos = 0;					// set cursor to 0

	// open file
	f = fopen(filename, "rb");
	if (!f)
	{
		free(file);
		return NULL;
	}

	// get size of file
	fseek(f, 0, SEEK_END);
	fgetpos(f, &s);
	fseek(f, 0, SEEK_SET);
	file->size = (long)s;


	// and load it
	file->buf =(char*)malloc((unsigned)s);
	if (!file->buf)
	{
		fclose(f);
		free(file);
		return NULL;
	}

	fread( file->buf, (unsigned)s, 1, f );
	fclose(f);

	return file;
} // _file_loadfromdisk




// "loads" a file from the pak into a buffer
// actually only copies the file from the pak buffer into another

file_t *_file_loadfrompak			// returns a pointer to the buffer
	( 
		char *filename 				// (in)  name of the file to load
	)
{
	char	*b = NULL;				// the buffer
	file_t	*file;					// the file descriptor

	file = (file_t*)malloc(sizeof(file_t));
	if (!file) return NULL;

	file->pos = 0;					// set cursor to 0

	if (!_pak_file)	// no pak no fun
	{
		free(file);
		return NULL;
	}

	// check every file in pak
	for (int i=0; i<_pak_filecount; i++)
	{
		// check if names are identical
		if (!_strcmpi( filename, _pak_header[i].filename ))
		{
			// save size
			file->size = _pak_header[i].size;

			// create buffer and "load" file
			file->buf = (char*)malloc(_pak_header[i].size);

			if (!file->buf)
			{
				free(file);
				return NULL;
			}

			// file is at beginning of _pak_header + offset
			memcpy(file->buf, _pak_file->buf+_pak_header[i].offset, _pak_header[i].size);

			return file;
		} // if filename == pakfilename[i]
	} // for every pakfile

	free(file);
	return NULL;					// file will be NULL here anyway
} // _file_loadfrompak




// _the_ file load function. tries to read a file from disk
// if it fails it tries to read it from a pak

file_t *_file_load					// returns a pointer to the buffer
	( 
		char *filename				// (in)  name of the file to load
	)
{
	file_t	*f;

	f = _file_loadfromdisk( filename );

	if (f) return f;

	return _file_loadfrompak( filename );
} // _file_load



// loads a pak pak into memory

int _file_loadpak					// returns nonzero if successful 0 otherwise
	( 
		char *filename				// the filename of the pak
	)
{
	_file_closepak();

	// try to load
	_pak_file = _file_loadfromdisk( filename );

	// failed?
	if (!_pak_file) return 0;

	memcpy(&_pak_filecount, _pak_file->buf, sizeof(_pak_filecount));

	// header is beginning of the buffer + size of the file count
	_pak_header = (fileentry_t*)((long)_pak_file->buf+sizeof(_pak_filecount));

	return 1;
} // _file_loadpak



// closes a file
void _file_close( file_t *file )
{
	if (!file) return;

	if (file->buf) free(file->buf);
	free(file);
} // _file_close



// frees the memory from the pak
void _file_closepak()
{
	_file_close( _pak_file );
} // _file_closepak



// do a seek

int _file_seek						// returns the read int
	( 
		file_t *f,					// file descriptor
		long offset,
		int	origin
	)
{
	if (!f || !f->buf)
		return 0;

	switch(origin)
	{
		case SEEK_CUR: f->pos += offset; break;
		case SEEK_END: f->pos = f->size-offset; break;
		case SEEK_SET: f->pos = offset; break;
		default: return 0;
	}

	if (f->pos<0 || f->pos>f->size || !f->buf)
		return 0;

	return 1;
} // _file_seek




// returns wheter a file is at its end

int _file_eof					// nonzero if at end 0 otherwise
	( 
		file_t *file			// the file descriptor
	)
{
	return file->pos>=file->size;
} // _fgetc_eof




// --------------------------------------------------------------------------
// buffer access functions below


// reads a char from a buffer

int _file_getc					// returns the read char as an int
	( 
		file_t *file			// the file descriptor
	)
{
	int		c;

	c = *(file->buf+file->pos);
	file->pos++;
	return c;
} // _file_getc



// reads an int from a buffer

int _file_geti						// returns the read int
	( 
		file_t *file				// the file descriptor
	)
{
	int		c;

	c = _file_getc(file);
	c |= _file_getc(file)<<8;

	return c;

} // _file_geti


// read bytes at ptr as bpp-bit pixel

void _file_getpixel
	( 
		file_t *file,				// the file descriptor
		char *r,					// pointer to where to store the red value to
		char *g,					// pointer to where to store the green value to 
		char *b,					// pointer to where to store the blue value to 
		char *a, 					// pointer to where to store the alpha value to
		char bpp					// bits per pixel
	)
{
	switch(bpp)
	{
		case 32:					// 32 bits are easy
			*r = _file_getc( file );
			*g = _file_getc( file );
			*b = _file_getc( file );
			*a = _file_getc( file );
			break;

		case 24:					// 24 too
			*r = _file_getc( file );
			*g = _file_getc( file );
			*b = _file_getc( file );
			*a = (char)255;
			break;

		case 16:					// what the... ? ;)

			// msb in byte 1 indicates alpha state if its set
			// make pixel non transparent
			
			char c1,c2;

			c1 = _file_getc( file );
			c2 = _file_getc( file );

			*a = c1&0x80?255:0;

			*r = (c1&124)<<1;
			*g = (c1&3<<6) | (c2&224)>>2;
			*b = (c2&31)<<3;
			break;
	}

} // _file_getpixel



// read a chunk of data from a buf
int	_file_read						// returns number of items read
	( 
		char *out,					// buffer to write to
		long size,					// size of items to read
		int n,						// number of items to read
		file_t *file				// file descriptor
	)
{
	for (int i=0; i<n; i++)
	{
		// FIXME: eof!
/*
		if (file->pos + size >= file->size)
			;//break;					// eof :)
*/
		memcpy( out+size*i, file->buf+file->pos, size );
		file->pos += size;
	}

	return n;
} // _file_read

// ----------------------------------------------------------------------- //
